//Author: Fabienne JEZEQUEL, contact:  Fabienne.Jezequel@lip6.fr
//example of execution command:
//matmul_cad 1000 3
#include <math.h>
#include <stdio.h>
#include <cadna.h>
#include <sys/time.h>

float random_float()
{
  float r = (float)(rand()-RAND_MAX/2.)/ (float)(RAND_MAX/2.);
  int c=rand()%4;
  r=r*pow(10.,c);
  return r;
}

void mat_multiply_float_st(float_st *a, float_st *b, float_st *c, int Dim){
   float_st aux;
   int i, j,k;
    for (i=0;i<Dim;i++){
      for (j=0;j<Dim;j++){
             aux = 0.0;
             for (k=0;k<Dim;k++)
                   aux = aux + a[i*Dim+k]*b[k*Dim+j];
            c[i*Dim+j] = aux;
      }
  }
}
    
void mat_multiply_float(float *a, float *b, float *c, int Dim){
   float aux;
   int i, j,k;
    for (i=0;i<Dim;i++){
      for (j=0;j<Dim;j++){
             aux = 0.0;
             for (k=0;k<Dim;k++)
                   aux = aux + a[i*Dim+k]*b[k*Dim+j];
            c[i*Dim+j] = aux;
      }
  }
}

int main(int argc, char *argv[]){
  int Dim = atoi(argv[1]);
  int per= atoi(argv[2]);
  cadna_init(0); //-1);
  struct timeval begin, end, begin2, end2;
  double elapsed, elapsed_without, elapsed_compute_without;//, ratio;

  float_st *a_st;
  float_st *b_st;
  float_st *c_st; 
  float_st aux_st;
  float_st *c; 
  
  float *a_x; 
  float *b_x; 
  float *c_x; 
  float *a_y;
  float *b_y; 
  float *c_y; 
  float *a_z;
  float *b_z;
  float *c_z; 
  int i, j, k;
  float aux;
  
  a_st  = (float_st *)calloc(Dim*Dim,sizeof(float_st));
  b_st  = (float_st *)calloc(Dim*Dim,sizeof(float_st));
  c_st  = (float_st *)calloc(Dim*Dim,sizeof(float_st));
  c  = (float_st *)calloc(Dim*Dim,sizeof(float_st));
  
  a_x  = (float *)calloc(Dim*Dim,sizeof(float));
  b_x  = (float *)calloc(Dim*Dim,sizeof(float));
  c_x  = (float *)calloc(Dim*Dim,sizeof(float));

  a_y  = (float *)calloc(Dim*Dim,sizeof(float));
  b_y   =  (float *)calloc(Dim*Dim,sizeof(float));
  c_y  =  (float *)calloc(Dim*Dim,sizeof(float));
  
  a_z  = (float *)calloc(Dim*Dim,sizeof(float));
  b_z  = (float *)calloc(Dim*Dim,sizeof(float));
  c_z  = (float *)calloc(Dim*Dim,sizeof(float));

  double exponent=(double)(-per);
  for(i=0;i<Dim*Dim;i++){
      aux_st=random_float();
      aux_st.data_st(pow(10.,exponent),0) ; //rel error;
      a_st[i] =  aux_st;
      aux_st=random_float();
      aux_st.data_st(pow(10.,exponent),0) ; //rel error;
      b_st[i] =  aux_st;    
  }
  /*
  for (i=0;i<5;i++){
    printf("a_st(%2d) = %s, acc=%2d \n",i,strp(a_st[i]),a_st[i].nb_significant_digit());    
  }
  printf("\n");
  */
  //start:
  gettimeofday(&begin, NULL);
  mat_multiply_float_st(a_st,b_st,c_st,Dim);
  //stop:
  gettimeofday(&end, NULL);
  elapsed = (end.tv_sec - begin.tv_sec)
    +  ((end.tv_usec - begin.tv_usec)/1000000.0);
  printf("time CADNA %9.2e\n",elapsed);
 
  int *acc_st,  *acc ;
  acc_st = (int *)calloc(Dim*Dim,sizeof(int));
  acc = (int *)calloc(Dim*Dim,sizeof(int));
  for(i=0;i<Dim*Dim;i++)
    acc_st[i]=c_st[i].nb_significant_digit();
  /*
  for (i=0;i<5;i++){
    printf("c_st(%2d) = %s, acc=%2d \n",i,strp(c_st[i]),c_st[i].nb_significant_digit());    
  }
  printf("\n");
  */
  //start
  gettimeofday(&begin, NULL);
  for(i=0;i<Dim*Dim;i++){
          a_x[i] =   a_st[i].getx();
	  a_y[i] =   a_st[i].gety();
	  a_z[i] =   a_st[i].getz();
          b_x[i] =   b_st[i].getx();
	  b_y[i] =   b_st[i].gety();
	  b_z[i] =   b_st[i].getz(); 
  }
  gettimeofday(&begin2, NULL);
  mat_multiply_float(a_x,b_x,c_x,Dim);
  mat_multiply_float(a_y,b_y,c_y,Dim);
  mat_multiply_float(a_z,b_z,c_z,Dim);
  gettimeofday(&end2, NULL);
  for(i=0;i<Dim*Dim;i++){
    c[i]=float_st(c_x[i], c_y[i], c_z[i]);
  }
   //stop
  gettimeofday(&end, NULL);
  elapsed_compute_without = (end2.tv_sec - begin2.tv_sec)
    +  ((end2.tv_usec - begin2.tv_usec)/1000000.0);
  elapsed_without = (end.tv_sec - begin.tv_sec)
    +  ((end.tv_usec - begin.tv_usec)/1000000.0);
  printf("complete time without CADNA %9.2e\n",elapsed_without);
  printf("ratio: %8.1e\n",elapsed/elapsed_without);
  printf("compute time without CADNA %9.2e\n",elapsed_compute_without);
  printf("ratio: %8.1e\n",elapsed/elapsed_compute_without);
  free(a_st);  
  free(b_st);
  free(   a_x );
  free(   b_x); 
  free(   c_x); 
   free(  a_y);
   free(  b_y );
   free( c_y);
   free(  a_z );
   free(  b_z); 
   free(  c_z );

   
  for(i=0;i<Dim*Dim;i++)
    acc[i]=c[i].nb_significant_digit();
  /*  for (i=0;i<5;i++){
    printf("c(%2d) = %s, acc=%2d \n",i,strp(c[i]),c[i].nb_significant_digit());    
  }
  printf("\n");
  */
 
  
  float *diff;  
  diff = (float *)calloc(Dim*Dim,sizeof(float));
  float max_err=0,min_err=1; 
  float mean_err=0;
  float_st  c_st_max,c_max;
  for(i=0;i<Dim*Dim;i++){
    if (c_st[i] ==0)  printf("zero for c_st=%s c=%s\n",strp(c_st[i]) ,strp(c[i]) );
     diff[i]=fabs(   (c_st[i]-c[i])/ c_st[i] );
     // if (diff[i]>pow(10.,1-c_st[i].nb_significant_digit()))
     //  printf("diff=%9.2e c_st=%s c=%s\n", diff[i], strp(c_st[i]) ,strp(c[i]) );
     if  (diff[i]>max_err) {
                 max_err=diff[i];
		 c_st_max=c_st[i];
		 c_max=c[i];
     }
     if  (diff[i]<min_err) min_err=diff[i];
     mean_err=mean_err+ diff[i];
  }
  mean_err=mean_err/(Dim*Dim);
  printf("rel error:\nmean= %9.2e, max=%9.2e min=%9.2e\n",mean_err,max_err,min_err);
  printf("max for c_st=%s c=%s\n",strp(c_st_max) ,strp(c_max) );

  free( diff);

  int max_acc=0,min_acc=20; 
  double mean=0;
  for(i=0;i<Dim*Dim;i++){
     if  (acc_st[i]>max_acc) max_acc=acc_st[i];
     if  (acc_st[i]<min_acc) min_acc=acc_st[i];
     mean=mean+ acc_st[i];
  }
  mean=mean/(Dim*Dim);
  printf("With CADNA, acc= %9.2e %2d-%2d\n",mean,min_acc,max_acc);
  
  max_acc=0,min_acc=20; 
  mean=0;
  for(i=0;i<Dim*Dim;i++){
     if  (acc[i]>max_acc) max_acc=acc[i];
     if  (acc[i]<min_acc) min_acc=acc[i];
     mean=mean+ acc[i];
  }
  mean=mean/(Dim*Dim);
  printf("Without CADNA, acc= %9.2e %2d-%2d\n",mean,min_acc,max_acc);

    
  int *diff_acc;  
  diff_acc = (int *)calloc(Dim*Dim,sizeof(int));
  int max=0,min=20; 
  mean=0;
  for(i=0;i<Dim*Dim;i++){
     diff_acc[i]=fabs(acc_st[i]-acc[i]);
     //     if (diff_acc[i] >=2) printf("c_st=%s c=%s\n",strp(c_st[i]) ,strp(c[i]) );
     if  (diff_acc[i]>max) max=diff_acc[i];
     if  (diff_acc[i]<min) min=diff_acc[i];
     mean=mean+ diff_acc[i];
  }
  mean=mean/(Dim*Dim);
  printf("diff accuracy:\nmean= %9.2e, max=%2d min=%2d\n",mean,max,min);

  cadna_end();
  free(  c_st );
  free(  c);
   free( acc_st);
   free(  acc);  
   free(  diff_acc);  
 
  return 1;
}










