//Author: Fabienne JEZEQUEL, contact:  Fabienne.Jezequel@lip6.fr
//example of execution command:
//matvecmul_cad 1000 3
#include <math.h>
#include <stdio.h>
#include <cadna.h>

float random_float()
{
  float r = (float)(rand()-RAND_MAX/2.)/ (float)(RAND_MAX/2.);
  int c=rand()%4;
  r=r*pow(10.,c);
  return r;
}

void mat_vec_multiply_float_st(float_st *a, float_st *x, float_st *y, int Dim){
   float_st aux;
   int i, j;
    for (i=0;i<Dim;i++){
    aux = 0.0;
    for (j=0;j<Dim;j++)
      aux = aux + a[i*Dim+j]*x[j];
    y[i] = aux;
  }
}

void mat_vec_multiply_float(float *a, float *x, float *y, int Dim){
   float aux;
   int i, j;
    for (i=0;i<Dim;i++){
    aux = 0.0;
    for (j=0;j<Dim;j++)
      aux = aux + a[i*Dim+j]*x[j];
    y[i] = aux;
  }
}

int main(int argc, char *argv[]){
  int Dim = atoi(argv[1]);
  int per= atoi(argv[2]);
  cadna_init(-1);

  float_st *a_st; 
  float_st *x_st; 
  float_st *y_st; 
  float_st *y;
  float_st aux_st;
  
  float *a_x; 
  float *x_x; 
  float *y_x; 
  float *a_y;
  float *x_y; 
  float *y_y; 
  float *a_z; 
  float *x_z; 
  float *y_z; 
  
  int i, j, k;
  float aux;
  
  a_st  = (float_st *)calloc(Dim*Dim,sizeof(float_st));
  x_st   = (float_st *)calloc(Dim,sizeof(float_st));
  y_st  = (float_st *)calloc(Dim,sizeof(float_st));
  y  = (float_st *)calloc(Dim,sizeof(float_st));
    
  a_x  = (float *)calloc(Dim*Dim,sizeof(float));
  x_x   = (float *)calloc(Dim,sizeof(float));
  y_x  = (float *)calloc(Dim,sizeof(float));
  
  a_y  = (float *)calloc(Dim*Dim,sizeof(float));
  x_y   = (float *)calloc(Dim,sizeof(float));
  y_y  = (float *)calloc(Dim,sizeof(float));
  
  a_z  = (float *)calloc(Dim*Dim,sizeof(float));
  x_z   = (float *)calloc(Dim,sizeof(float));
  y_z  = (float *)calloc(Dim,sizeof(float));  

  float exponent=(float)(-per);
  for(i=0;i<Dim;i++){
    for (j=0;j<Dim;j++){
      aux_st=random_float();
      aux_st.data_st(pow(10.,exponent),0) ; //rel error;
      a_st[i*Dim+j] =  aux_st;
    }
    aux_st=random_float();
    aux_st.data_st(pow(10.,exponent),0) ; //rel error;
    x_st[i]=aux_st;
  }

  for (i=0;i<5;i++){
    printf("x_st(%2d) = %s, acc=%2d \n",i,strp(x_st[i]),x_st[i].nb_significant_digit());    
  }
  printf("\n");
  mat_vec_multiply_float_st(a_st,x_st,y_st,Dim);
  
  int *acc_st,   *acc ,*diff_acc;  
  acc_st = (int *)calloc(Dim,sizeof(int));
  acc = (int *)calloc(Dim,sizeof(int));
  diff_acc = (int *)calloc(Dim,sizeof(int));
  
  for(i=0;i<Dim;i++)
    acc_st[i]=y_st[i].nb_significant_digit();
  for (i=0;i<5;i++){
    printf("y_st(%2d) = %s, acc=%2d \n",i,strp(y_st[i]),y_st[i].nb_significant_digit());    
  }
  printf("\n");
  
  for(i=0;i<Dim;i++){
    for (j=0;j<Dim;j++){
          a_x[j*Dim+i] =   a_st[j*Dim+i].getx();
	  a_y[j*Dim+i] =   a_st[j*Dim+i].gety();
	  a_z[j*Dim+i] =   a_st[j*Dim+i].getz(); 
    }
    
    x_x[i]= x_st[i].getx();
    x_y[i]= x_st[i].gety();
    x_z[i]= x_st[i].getz();
  }
  
  free(a_st);  
  free(x_st);

  mat_vec_multiply_float(a_x,x_x,y_x,Dim);
  mat_vec_multiply_float(a_y,x_y,y_y,Dim);
  mat_vec_multiply_float(a_z,x_z,y_z,Dim);

  for(i=0;i<Dim;i++){
    y[i]=float_st(y_x[i], y_y[i], y_z[i]);
  }
  
  for(i=0;i<Dim;i++)
    acc[i]=y[i].nb_significant_digit();
  for (i=0;i<5;i++){
    printf("y(%2d) = %s, acc=%2d \n",i,strp(y[i]),y[i].nb_significant_digit());    
  }
  printf("\n");
  
  float *diff;  
  diff = (float *)calloc(Dim,sizeof(float));
  float max_err=0,min_err=1; 
  float mean_err=0;
  float_st  y_st_max, y_max;
  for(i=0;i<Dim;i++){
     if (y_st[i] ==0)  printf("zero for y_st=%s y=%s\n",strp(y_st[i]) ,strp(y[i]) );
     diff[i]=fabs( (y_st[i]-y[i])/ y_st[i] );
     if (diff[i]>pow(10.,1-y_st[i].nb_significant_digit()))
       printf("diff=%9.2e y_st=%s y=%s\n", diff[i], strp(y_st[i]) ,strp(y[i]) );
     if  (diff[i]>max_err) {
                 max_err=diff[i];
		 y_st_max=y_st[i];
		 y_max=y[i];
     }
     if  (diff[i]<min_err) min_err=diff[i];
     mean_err=mean_err+ diff[i];
  }
  mean_err=mean_err/Dim;
  printf("rel error:\nmean= %9.2e, max=%9.2e min=%9.2e\n",mean_err,max_err,min_err);
  printf("max for y_st=%s y=%s\n",strp(y_st_max) ,strp(y_max) );
  free(  diff);

  int max_acc=0,min_acc=20; 
  float mean=0;
  for(i=0;i<Dim;i++){
     if  (acc_st[i]>max_acc) max_acc=acc_st[i];
     if  (acc_st[i]<min_acc) min_acc=acc_st[i];
     mean=mean+ acc_st[i];
  }
  mean=mean/Dim;
  printf("With CADNA, acc= %9.2e %2d-%2d\n",mean,min_acc,max_acc);
  
  max_acc=0,min_acc=20; 
  mean=0;
  for(i=0;i<Dim;i++){
     if  (acc[i]>max_acc) max_acc=acc[i];
     if  (acc[i]<min_acc) min_acc=acc[i];
     mean=mean+ acc[i];
  }
  mean=mean/Dim;
  printf("Without CADNA, acc= %9.2e %2d-%2d\n",mean,min_acc,max_acc);
     
  max_acc=0,min_acc=20; 
  mean=0;
  for(i=0;i<Dim;i++){
     diff_acc[i]=fabs(acc_st[i]-acc[i]);
     if  (diff_acc[i]>max_acc) max_acc=diff_acc[i];
     if  (diff_acc[i]<min_acc) min_acc=diff_acc[i];
     mean=mean+ diff_acc[i];
  }
  mean=mean/Dim;
  printf("mean acc diff= %9.2e, max=%2d min=%2d\n",mean,max_acc,min_acc);

  cadna_end();

  free(  y);
  free(  y_st );
   free( acc_st);
   free(  acc);  
   free(  diff_acc);  
  
  free(   a_x );
  free(   x_x); 
  free(   y_x); 
  
   free(  a_y);
   free(  x_y );
    free( y_y); 

   free(  a_z );
   free(  x_z); 
   free(  y_z );
  
  return 1;
}














