//Author: Fabienne JEZEQUEL, contact:  Fabienne.Jezequel@lip6.fr
//example of execution command:
//matvecmul_cad_double_MKL 1000 10
#include <math.h>
#include <stdio.h>
#include <cadna.h>
#include <mkl.h>
#include <sys/time.h>

double random_double()
{
  double r = (double)(rand()-RAND_MAX/2.)/ (double)(RAND_MAX/2.);
  int c=rand()%20;
  r=r*pow(10.,c);
  return r;
}

void mat_vec_multiply_double_st(double_st *a, double_st *x, double_st *y, int Dim){
   double_st aux;
   int i, j;
    for (i=0;i<Dim;i++){
    aux = 0.0;
    for (j=0;j<Dim;j++)
      aux = aux + a[i*Dim+j]*x[j];
    y[i] = aux;
  }
}

void mat_vec_multiply_double(double *a, double *x, double *y, int Dim){
   double aux;
   int i, j;
    for (i=0;i<Dim;i++){
    aux = 0.0;
    for (j=0;j<Dim;j++)
      aux = aux + a[i*Dim+j]*x[j];
    y[i] = aux;
  }
}

void mat_vec_multiply_double_MKL(double *a, double *x, double *y, int Dim){

  cblas_dgemv(CblasRowMajor, CblasNoTrans,Dim,Dim,1.,a,Dim,x,1,0.,y,1);

  /*  
   double aux;
   int i, j;
    for (i=0;i<Dim;i++){
    aux = 0.0;
    for (j=0;j<Dim;j++)
      aux = aux + a[i*Dim+j]*x[j];
    y[i] = aux;
  }
  */
}


int main(int argc, char *argv[]){
  int Dim = atoi(argv[1]);
  int per= atoi(argv[2]);
  struct timeval begin, end, begin2, end2;
  double elapsed, elapsed_without, elapsed_compute_without;
  cadna_init(0); //-1);

  double_st *a_st; 
  double_st *x_st; 
  double_st *y_st; 
  double_st *y;
  double_st aux_st;
  
  double *a_x; 
  double *x_x; 
  double *y_x;
  double *a_y;
  double *x_y; 
  double *y_y; 
  double *a_z; 
  double *x_z; 
  double *y_z; 
  
  int i, j, k;
  double aux;
  
  a_st  = (double_st *)calloc(Dim*Dim,sizeof(double_st));
  x_st   = (double_st *)calloc(Dim,sizeof(double_st));
  y_st  = (double_st *)calloc(Dim,sizeof(double_st));
  y  = (double_st *)calloc(Dim,sizeof(double_st));
    
  a_x  = (double *)calloc(Dim*Dim,sizeof(double));
  x_x   = (double *)calloc(Dim,sizeof(double));
  y_x  = (double *)calloc(Dim,sizeof(double));
  
  a_y  = (double *)calloc(Dim*Dim,sizeof(double));
  x_y   = (double *)calloc(Dim,sizeof(double));
  y_y  = (double *)calloc(Dim,sizeof(double));
  
  a_z  = (double *)calloc(Dim*Dim,sizeof(double));
  x_z   = (double *)calloc(Dim,sizeof(double));
  y_z  = (double *)calloc(Dim,sizeof(double));  

  double exponent=(double)(-per);
  for(i=0;i<Dim;i++){
    for (j=0;j<Dim;j++){
      aux_st=random_double();
      aux_st.data_st(pow(10.,exponent),0) ; //rel error;
      a_st[i*Dim+j] =  aux_st;
    }
    aux_st=random_double();
    aux_st.data_st(pow(10.,exponent),0) ; //rel error;
    x_st[i]=aux_st;
  }

  /*
  for (i=0;i<5;i++){
    printf("x_st(%2d) = %s, acc=%2d \n",i,strp(x_st[i]),x_st[i].nb_significant_digit());    
  }
  printf("\n");
  */

  
  //start
  gettimeofday(&begin, NULL);
  mat_vec_multiply_double_st(a_st,x_st,y_st,Dim);
  //stop
  gettimeofday(&end, NULL);
  elapsed = (end.tv_sec - begin.tv_sec)
    +  ((end.tv_usec - begin.tv_usec)/1000000.0);
  printf("time CADNA %9.2e\n",elapsed);
  
  int *acc_st,   *acc ,*diff_acc;  
  acc_st = (int *)calloc(Dim,sizeof(int));
  acc = (int *)calloc(Dim,sizeof(int));
  diff_acc = (int *)calloc(Dim,sizeof(int));
  
  for(i=0;i<Dim;i++)
    acc_st[i]=y_st[i].nb_significant_digit();
  /*
  for (i=0;i<5;i++){
    printf("y_st(%2d) = %s, acc=%2d \n",i,strp(y_st[i]),y_st[i].nb_significant_digit());    
  }
  printf("\n");
  */
   //start
  gettimeofday(&begin, NULL);
  for(i=0;i<Dim*Dim;i++){
          a_x[i] =   a_st[i].getx();
	  a_y[i] =   a_st[i].gety();
	  a_z[i] =   a_st[i].getz(); 
    }
for(i=0;i<Dim;i++){
    x_x[i]= x_st[i].getx();
    x_y[i]= x_st[i].gety();
    x_z[i]= x_st[i].getz();
  }
  gettimeofday(&begin2, NULL);
  /*
  mat_vec_multiply_double(a_x,x_x,y_x,Dim);
  mat_vec_multiply_double(a_y,x_y,y_y,Dim);
  mat_vec_multiply_double(a_z,x_z,y_z,Dim);
  */
  mat_vec_multiply_double_MKL(a_x,x_x,y_x,Dim);
  mat_vec_multiply_double_MKL(a_y,x_y,y_y,Dim);
  mat_vec_multiply_double_MKL(a_z,x_z,y_z,Dim);
gettimeofday(&end2, NULL);
  
  for(i=0;i<Dim;i++){
    y[i]=double_st(y_x[i], y_y[i], y_z[i]);
  }
  //stop
  gettimeofday(&end, NULL);
  elapsed_compute_without = (end2.tv_sec - begin2.tv_sec)
    +  ((end2.tv_usec - begin2.tv_usec)/1000000.0);
  elapsed_without = (end.tv_sec - begin.tv_sec)
    +  ((end.tv_usec - begin.tv_usec)/1000000.0);
  printf("complete time without CADNA %9.2e\n",elapsed_without);
  printf("ratio: %8.1e\n",elapsed/elapsed_without);
  printf("compute time without CADNA %9.2e\n",elapsed_compute_without);
  printf("ratio: %8.1e\n",elapsed/elapsed_compute_without);
  free(a_st);  
  free(x_st);
  
  for(i=0;i<Dim;i++)
    acc[i]=y[i].nb_significant_digit();
  for (i=0;i<5;i++){
    printf("y(%2d) = %s, acc=%2d \n",i,strp(y[i]),y[i].nb_significant_digit());    
  }
  printf("\n");
  
  double *diff;  
  diff = (double *)calloc(Dim,sizeof(double));
  double max_err=0,min_err=1; 
  double mean_err=0;
  double_st  y_st_max, y_max;
  for(i=0;i<Dim;i++){
     if (y_st[i] ==0)  printf("zero for y_st=%s y=%s\n",strp(y_st[i]) ,strp(y[i]) );
     diff[i]=fabs( (y_st[i]-y[i])/ y_st[i] );
     if (diff[i]>pow(10.,1-y_st[i].nb_significant_digit()))
       printf("diff=%9.2e y_st=%s y=%s\n", diff[i], strp(y_st[i]) ,strp(y[i]) );
     if  (diff[i]>max_err) {
                 max_err=diff[i];
		 y_st_max=y_st[i];
		 y_max=y[i];
     }
     if  (diff[i]<min_err) min_err=diff[i];
     mean_err=mean_err+ diff[i];
  }
  mean_err=mean_err/Dim;
  printf("rel error:\nmean= %9.2e, max=%9.2e min=%9.2e\n",mean_err,max_err,min_err);
  printf("max for y_st=%s y=%s\n",strp(y_st_max) ,strp(y_max) );
  free(  diff);

  int max_acc=0,min_acc=20; 
  double mean=0;
  for(i=0;i<Dim;i++){
     if  (acc_st[i]>max_acc) max_acc=acc_st[i];
     if  (acc_st[i]<min_acc) min_acc=acc_st[i];
     mean=mean+ acc_st[i];
  }
  mean=mean/Dim;
  printf("With CADNA, acc= %9.2e %2d-%2d\n",mean,min_acc,max_acc);
  
  max_acc=0,min_acc=20; 
  mean=0;
  for(i=0;i<Dim;i++){
     if  (acc[i]>max_acc) max_acc=acc[i];
     if  (acc[i]<min_acc) min_acc=acc[i];
     mean=mean+ acc[i];
  }
  mean=mean/Dim;
  printf("Without CADNA, acc= %9.2e %2d-%2d\n",mean,min_acc,max_acc);
     
  max_acc=0,min_acc=20; 
  mean=0;
  for(i=0;i<Dim;i++){
     diff_acc[i]=fabs(acc_st[i]-acc[i]);
     if  (diff_acc[i]>max_acc) max_acc=diff_acc[i];
     if  (diff_acc[i]<min_acc) min_acc=diff_acc[i];
     mean=mean+ diff_acc[i];
  }
  mean=mean/Dim;
  printf("mean acc diff= %9.2e, max=%2d min=%2d\n",mean,max_acc,min_acc);

  cadna_end();

  free(  y);
  free(  y_st );
   free( acc_st);
   free(  acc);  
   free(  diff_acc);  
  
  free(   a_x );
  free(   x_x); 
  free(   y_x); 
  
   free(  a_y);
   free(  x_y );
    free( y_y); 

   free(  a_z );
   free(  x_z); 
   free(  y_z );
  
  return 1;
}














